#include "SDPdef.h"

int CholFactor ( double **mat, int p, double *temp)
{ /* Given the upper triangular half of a symmetric matrix A,
     this routine finds the Cholesky factorization A = LDL^T 
     From Golub and VanLoan.                                */
  
  int i,j,k;
  double t;

  for (j=0;j<p;j++){
    for (i=0;i<j;i++){
      temp[i] = mat[i][j-i]*mat[i][0];
    }
    temp[j] = mat[j][0];
    for (k=0;k<j;k++){
      temp[j] -= mat[k][j-k] * temp[k];
    }
    if (temp[j] < 1e-16){
      /*     printf("Very small \n"); */
      return (j+1);
    }
    mat[j][0] = temp[j];
    for ( k=j+1; k<p; k++ ){
      t = 0.0;
      for (i=0;i<j;i++){
	t += mat[i][k-i]*temp[i];
      }
      mat[j][k-j] = (mat[j][k-j] - t)/temp[j];
    }
  }
  return 0;
}


void CholSolver (double **ch, double *ans, double *rhs, int p)
{ /* Given an LDL^T decomposition of a matrix A, this 
    solves A (ans) = rhs */

  int i,j;
  
  /* Forward Substitution */
  for (i=0;i<p;i++){
    ans[i] = rhs[i];
    for (j=0;j<i;j++){
      ans[i] -=  ans[j]*ch[j][i-j];
    }
  }

  /* Diagonal Substitution */
  for (i=0; i<p; i++){
    ans[i] = ans[i] / ch[i][0];
  }

  /* Back Substitution */
  for (i=p-1;i>=0; i--){
    for (j=i+1;j<p; j++){
      ans[i] -= ch[i][j-i] * ans[j];
    }
  }

}


void LUFactor(double **mat, int dim){
  int i,j,k;

  for (k=0; k<dim-1; k++){
    for (i=k+1; i<dim; i++){
      mat[i][k]=mat[i][k]/mat[k][k];
    }
    for (i=k+1; i<dim; i++){
      for (j=k+1; j<dim; j++){
	mat[i][j] = mat[i][j] - mat[i][k]*mat[k][j];
      }
    }
  }
}

void LUSolver(double **mat, double *v, int dim){
  int i,j;
  double temp;

  /* Forward Substitution */
  for (i=0;i<dim;i++){
    for (j=0;j<i;j++){
      v[i] = v[i]- mat[i][j]*v[j];
    }
  }

  /* Back Substitution */
  for (i=dim-1;i>=0; i--){
    temp = 0.0;
    for (j=i+1;j<dim; j++){
      temp = temp + mat[i][j] * v[j];
    }
    v[i] = (v[i] - temp)/mat[i][i];
  }
}



void MakeSMWmat( sdpdat *sdt, double *co){

  int i,j,nden;
  smwTools *smw=sdt->smw;

  nden=smw->ndense;

  for (i=0; i<nden; i++){
    MakeDense(&(sdt->amats->rows[smw->dind[i]]),sdt->wd,sdt->n);
    ChlSolve(sdt->sf, sdt->wd, smw->upd[i]);
    smw->coef[i]=-co[smw->dind[i]];
    for (j=0; j< nden; j++){
      smw->mat[j][i] = VDotSD(&(sdt->amats->rows[smw->dind[j]]), smw->upd[i])
	  *(smw->coef[i]); 
      }
  }

  for (i=0;i<nden;i++){
    smw->mat[i][i] += 1.0;
  }

  LUFactor(smw->mat, nden);
}


void SMWupdate(smwTools *smw, spvec *sv, double *dv, 
	       double *ans, int n, int spars){

  int j,k,nden = smw->ndense;

  if (spars == TRUE){
    for (k=0;k<nden;k++) {
      smw->v1[k] = VDotSD(sv, smw->upd[k]);
    }
  }
  else {
    for (k=0;k<nden;k++) {
      smw->v1[k] = Dot(dv, smw->upd[k],n);
    }
  }

  LUSolver(smw->mat, smw->v1, nden);

  for (k=0;k<nden;k++){
    smw->v1[k]= smw->v1[k]*(smw->coef[k]);
  }
  for (k=0;k<n;k++){
    for (j=0; j<nden; j++){
      ans[k] = ans[k] - smw->v1[j]*smw->upd[j][k];
    }
  }

}













